;Protected mode tutorial - scritto da Albe per la UIC
;



;lunghezza del body in settori (ci serve per sapere quanti settori
;deve caricare il bootsector)
%define BODY_SEC_LENGTH ((body_end-body) / 512) + 1

;====================================================================
;codice a 16 bit

	ORG	0x7C00		;0x7C00  dove ci troviamo all'inizio del bootsector
	BITS	16

;--------------------------------------------------------------------
;bootsector - qui carichiamo il resto del programma

boot:

;prima cosa da fare: inizializzare i registri di segmento e lo stack
	xor	ax,ax
	mov	ds,ax
	mov	es,ax
	mov	ss,ax
	mov	sp,0x7C00	;mettiamo lo stack appena sotto al bootsector

;ora carichiamo il resto del programma
	mov	ax,0x0200 + BODY_SEC_LENGTH	;al=numero di settori da leggere
	mov	cx,2		;iniziamo a leggere dal secondo settore
	xor	dx,dx
	mov	bx,0x7E00
	int	0x13

;saltiamo al body che abbiamo appena caricato
	jmp	0x0000:0x7E00

;--------------------------------------------------------------------
;bootsector filler

times (510-($-boot)) db 0	;riempiamo di zeri per paddare i 512 byte del bootsector
	dw	0xAA55		;bootsector magic number

;--------------------------------------------------------------------
;program body

body:

;salutiamo...
	push	word welcome_msg
	call	print

;1 - disabilitamo gli interrupt
	push	word cli_msg
	call	print
	cli
	
	in	al,0x70		;disabilito i NMI
	or	al,0x80
	out	0x70,al

;2 - abilitiamo la linea a20
	call	a20_enable

;3 - carichiamo la gdt
	push	word lgdt_msg
	call	print
	lgdt	[gdtr]

;4 - settiamo il bit PE
	push	word set_pmode
	call	print
	mov	eax,CR0
	or	al,1
	mov	CR0,eax

;5 - saltiamo al codice a 32 bit
	jmp	dword 8:body_32

;--------------------------------------------------------------------
;abilitazione della linea di indirizzamento a20

a20_enable:
;void a20_enable();

	push	ax

	push	word a20_msg
	call	print

	call	.kbd_wait_output

;leggiamo il current status del keyboard controller
	mov	al,0xD0		;0xD0 = read output port
	out	0x64,al
	
	call	.kbd_wait_input

	xor	ax,ax
	in	al,0x60		;current status in AL
	or	al,2		;settiamo il bit 2, che abilita la linea a20
	push	ax		;salviamolo per dopo
	
	call	.kbd_wait_output

;scriviamo lo status
	mov	al,0xD1		;0xD1 = write output port
	out	0x64,al
	
	call	.kbd_wait_output
	
	pop	ax
	out	0x60,al
	
	pop	ax
	
	ret

.kbd_wait_output:
;cicliamo finch il command buffer della tastiera non  vuoto
	in	al,0x64
	test	al,2
	jnz	short .kbd_wait_output
	ret

.kbd_wait_input:
;cicliamo finch non c' un dato disponibile nel buffer della tastiera
	in	al,0x64
	test	al,1
	jz	short .kbd_wait_input
	ret

;--------------------------------------------------------------------
;real mode print - scrive una stringa sullo schermo, usando l'int 10h

print:
;void print(char *message);

%define	message bp+4

	push	bp
	mov	bp,sp
	
	push	ax
	push	bx
	push	si
	
	mov	si,word [message]
	mov	ah,0x0E
	xor	bx,bx

.next_char:
	lodsb
	test	al,al
	jz	short .done
	int	0x10
	jmp	short .next_char

.done:
	pop	si
	pop	bx
	pop	ax
	
	mov	sp,bp
	pop	bp
	ret	2

%undef	message

;====================================================================
;codice a 32 bit

	BITS	32

body_32:
;inizializzazione dei selettori
	mov	ax,0x10
	mov	ds,ax
	mov	es,ax
	mov	ss,ax
	mov	esp,0x000A0000

;risalutiamo da pmode
	push	dword pmode_msg
	call	print_32

;contiamo la memoria
	call	mem_count

;the end
	push	dword hang_msg
	call	print_32
	
;	jmp	$		;hang
	cli
	hlt

;--------------------------------------------------------------------
;memory counter

mem_count:
	push	dword mem_count_msg
	call	print_32
	
	push	dword mem_count_result1_msg
	call	print_32
	
%define	COUNT_STEP 0x1000

	mov	esi,0x00100000-COUNT_STEP

.again:
	call	get_cursor_pos
	push	eax
	
	push	esi
	call	print_num_32
	
	call	set_cursor_pos

	mov	eax,0x1234ABCD
	mov	ebx,0xF1E2D3C4

	add	esi,COUNT_STEP
	
	mov	dword [esi],eax
	cmp	dword [esi],eax
	je	short .again
	mov	dword [esi],ebx
	cmp	dword [esi],ebx
	je	short .again
	
	push	esi
	call	print_num_32
	
	push	dword mem_count_result2_msg
	call	print_32
	
	ret

;--------------------------------------------------------------------
;routine vga a 32 bit - ho messo alcune routine di pi basso livello
;per la gestione del video in un file separato ("vga32.asm")

%include "vga32.asm"

;--------------------------------------------------------------------
;protected mode print - scrive una stringa direttamente nella memoria video

print_32:
;void print_32(char *message);
;

%define	message ebp+8

	push	ebp
	mov	ebp,esp
	
	push	eax
	push	esi
	
	mov	esi,dword [message]

.again:
	lodsb
	test	al,al
	jz	short .done
	push	dword 15
	push	eax
	call	print_char_32
	jmp	short .again

.done:
	pop	esi
	pop	eax
	
	mov	esp,ebp
	pop	ebp
	ret	4

%undef	message

;--------------------------------------------------------------------
;una routine per scrivere numeri hex

print_num_32:
;void print_num_32(int number);

%define	number	ebp+8

	push	ebp
	mov	ebp,esp
	
	push	eax
	push	ebx
	push	ecx
	
	mov	ecx,8
	mov	eax,dword [number]
	xor	ebx,ebx

.again:
	rol	eax,4
	mov	bl,al
	and	bl,0x0F
	add	bl,0x30
	cmp	bl,'9'
	jbe	.digit
	
	add	bl,7

.digit:
	push	dword 4
	push	ebx
	call	print_char_32
	
	loop	.again
	
	push	dword 4
	push	dword 'h'
	call	print_char_32
	
	pop	ecx
	pop	ebx
	pop	eax
	
	mov	esp,ebp
	pop	ebp
	ret	4

;====================================================================
;messaggi

welcome_msg:
	times 80 db '-'
	db	'            Protected mode tutorial -- scritto da Albe per la UIC',13,10
	times 80 db '-'
	db	13,10,13,10
	db	'UIC home page:     http://www.quequero.tk',13,10
	db	'                   http://www.scared.tk',13,10
	db	'                   http://quequero.cjb.net',13,10
	db	'La mia home page:  http://www.albe.tk',13,10
	db	'La mia email:      albe@bbk.org - commenti, critiche, dubbi sono benvenuti!',13,10,13,10,0

ok_msg:
	db	'ok',13,10,0

reboot_msg:
	db	'Premere un tasto per riavviare...',0

cpu_check_msg:
	db	'Controllo che la CPU sia 386+...',0
	
cpu_check_error:
	db	'386+ non trovato!',0

cli_msg:
	db	'Disabilito gli interrupt',13,10,0

a20_msg:
	db	'Abilito la linea A20',13,10,0

lgdt_msg:
	db	'Carico la GDT',13,10,0

set_pmode:
	db	'Setto PE...see you soon in protected mode!',13,10,13,10
	times	80 db '-'
	db	0

pmode_msg:
	db	10,'Rieccoci qua! Ma ora...siamo a 32 bit!',10,0

mem_count_msg:
	db	'Giusto per provare a fare qualcosa, contiamo la RAM...',10,0

mem_count_result1_msg:
	db	'Ho contato ',0

mem_count_result2_msg:
	db	' byte',10,0

hang_msg:
	db	'That',39,'s all folks! Ora...appendiamo la cornetta :)',10
	db	'Ci si vede su azzurra!',10,0


;--------------------------------------------------------------------
;GDT

GDTFLAG_G_4KB		equ 0x80
GDTFLAG_DB_32		equ 0x40
GDTFLAG_P		equ 0x80
GDTFLAG_DPL_0		equ 0
GDTFLAG_S		equ 0x10
GDTFLAG_TYPE_CODE	equ 0x0A
GDTFLAG_TYPE_DATA	equ 0x02

gdtr:
	dw	GDT_SIZE
	dd	null_desc

null_desc:
	times 8 db 0

code_desc:
	dw	0xFFFF		;limit 0-15
	dw	0x0000		;base 0-15
	db	0x00		;base 16-23
	db	GDTFLAG_P+GDTFLAG_S+GDTFLAG_TYPE_CODE
	db	GDTFLAG_G_4KB+GDTFLAG_DB_32+0x0F
	db	0x00		;base 24-31

data_desc:
	dw	0xFFFF		;limit 0-15
	dw	0x0000		;base 0-15
	db	0x00		;base 16-23
	db	GDTFLAG_P+GDTFLAG_S+GDTFLAG_TYPE_DATA
	db	GDTFLAG_G_4KB+GDTFLAG_DB_32+0x0F
	db	0x00		;base 24-31

GDT_SIZE equ	$ - null_desc

;====================================================================
;end marker
body_end:

;questo filler ci serve per costruire un'immagine vera e propria del floppy
;(cio, un'immagine da 1.4 MB). Bochs legge anche un'immagine incompleta,
;mentre VMVare vuole l'immagine da 1.4 MB (e probabilmente anche rawrite).
times	((18*2*80*512)-($-boot)) db 0
